# Stubs for errno

from typing import Mapping
import sys

errorcode = ...  # type: Mapping[int, str]

EPERM = ...  # type: int
ENOENT = ...  # type: int
ESRCH = ...  # type: int
EINTR = ...  # type: int
EIO = ...  # type: int
ENXIO = ...  # type: int
E2BIG = ...  # type: int
ENOEXEC = ...  # type: int
EBADF = ...  # type: int
ECHILD = ...  # type: int
EAGAIN = ...  # type: int
ENOMEM = ...  # type: int
EACCES = ...  # type: int
EFAULT = ...  # type: int
ENOTBLK = ...  # type: int
EBUSY = ...  # type: int
EEXIST = ...  # type: int
EXDEV = ...  # type: int
ENODEV = ...  # type: int
ENOTDIR = ...  # type: int
EISDIR = ...  # type: int
EINVAL = ...  # type: int
ENFILE = ...  # type: int
EMFILE = ...  # type: int
ENOTTY = ...  # type: int
ETXTBSY = ...  # type: int
EFBIG = ...  # type: int
ENOSPC = ...  # type: int
ESPIPE = ...  # type: int
EROFS = ...  # type: int
EMLINK = ...  # type: int
EPIPE = ...  # type: int
EDOM = ...  # type: int
ERANGE = ...  # type: int
EDEADLCK = ...  # type: int
ENAMETOOLONG = ...  # type: int
ENOLCK = ...  # type: int
ENOSYS = ...  # type: int
ENOTEMPTY = ...  # type: int
ELOOP = ...  # type: int
EWOULDBLOCK = ...  # type: int
ENOMSG = ...  # type: int
EIDRM = ...  # type: int
ECHRNG = ...  # type: int
EL2NSYNC = ...  # type: int
EL3HLT = ...  # type: int
EL3RST = ...  # type: int
ELNRNG = ...  # type: int
EUNATCH = ...  # type: int
ENOCSI = ...  # type: int
EL2HLT = ...  # type: int
EBADE = ...  # type: int
EBADR = ...  # type: int
EXFULL = ...  # type: int
ENOANO = ...  # type: int
EBADRQC = ...  # type: int
EBADSLT = ...  # type: int
EDEADLOCK = ...  # type: int
EBFONT = ...  # type: int
ENOSTR = ...  # type: int
ENODATA = ...  # type: int
ETIME = ...  # type: int
ENOSR = ...  # type: int
ENONET = ...  # type: int
ENOPKG = ...  # type: int
EREMOTE = ...  # type: int
ENOLINK = ...  # type: int
EADV = ...  # type: int
ESRMNT = ...  # type: int
ECOMM = ...  # type: int
EPROTO = ...  # type: int
EMULTIHOP = ...  # type: int
EDOTDOT = ...  # type: int
EBADMSG = ...  # type: int
EOVERFLOW = ...  # type: int
ENOTUNIQ = ...  # type: int
EBADFD = ...  # type: int
EREMCHG = ...  # type: int
ELIBACC = ...  # type: int
ELIBBAD = ...  # type: int
ELIBSCN = ...  # type: int
ELIBMAX = ...  # type: int
ELIBEXEC = ...  # type: int
EILSEQ = ...  # type: int
ERESTART = ...  # type: int
ESTRPIPE = ...  # type: int
EUSERS = ...  # type: int
ENOTSOCK = ...  # type: int
EDESTADDRREQ = ...  # type: int
EMSGSIZE = ...  # type: int
EPROTOTYPE = ...  # type: int
ENOPROTOOPT = ...  # type: int
EPROTONOSUPPORT = ...  # type: int
ESOCKTNOSUPPORT = ...  # type: int
ENOTSUP = ...  # type: int
EOPNOTSUPP = ...  # type: int
EPFNOSUPPORT = ...  # type: int
EAFNOSUPPORT = ...  # type: int
EADDRINUSE = ...  # type: int
EADDRNOTAVAIL = ...  # type: int
ENETDOWN = ...  # type: int
ENETUNREACH = ...  # type: int
ENETRESET = ...  # type: int
ECONNABORTED = ...  # type: int
ECONNRESET = ...  # type: int
ENOBUFS = ...  # type: int
EISCONN = ...  # type: int
ENOTCONN = ...  # type: int
ESHUTDOWN = ...  # type: int
ETOOMANYREFS = ...  # type: int
ETIMEDOUT = ...  # type: int
ECONNREFUSED = ...  # type: int
EHOSTDOWN = ...  # type: int
EHOSTUNREACH = ...  # type: int
EALREADY = ...  # type: int
EINPROGRESS = ...  # type: int
ESTALE = ...  # type: int
EUCLEAN = ...  # type: int
ENOTNAM = ...  # type: int
ENAVAIL = ...  # type: int
EISNAM = ...  # type: int
EREMOTEIO = ...  # type: int
EDQUOT = ...  # type: int
