from typing import Any
# from ..packages import six

# b = six.b

ACCEPT_ENCODING = ...  # type: Any

def make_headers(keep_alive=..., accept_encoding=..., user_agent=..., basic_auth=..., proxy_basic_auth=..., disable_cache=...): ...
